import typing
from enum import Enum
from dataclasses import dataclass, field

from .tokfmt import tokfmt, Token


@dataclass
class Value:
    """
    A unparsed list of tokens

    .. code-block:: c++

        int x = 0x1337;
                ~~~~~~
    """

    #: Tokens corresponding to the value
    tokens: typing.List[Token]

    def format(self) -> str:
        return tokfmt(self.tokens)


class AttributeStyle(Enum):
    """
    Indicates what style of attribute was found
    """

    #: ``[[deprecated]]``
    CXX = "cxx"

    #: ``__attribute__((deprecated))``
    GCC = "gcc"

    #: ``__declspec(deprecated)``
    MSVC = "msvc"


@dataclass
class Attribute:
    """
    An attribute

    .. code-block:: c++

        [[deprecated("message")]]
          ~~~~~~~~~  ~~~~~~~~~~~
    """

    #: Indicates what style of attribute this is
    style: AttributeStyle

    name: str

    #: Portion of the attribute inside the parens
    value: typing.Optional[Value] = None


@dataclass
class NamespaceAlias:
    """
    A namespace alias

    .. code-block:: c++

        namespace ANS = my::ns;
                  ~~~   ~~~~~~

    """

    alias: str

    #: These are the names (split by ::) for the namespace that this alias
    #: refers to, but does not include any parent namespace names. It may
    #: include a leading "::", but does not include a following :: string.
    names: typing.List[str]


@dataclass
class NamespaceDecl:
    """
    Namespace declarations

    .. code-block:: c++

        namespace foo::bar {}
                  ~~~~~~~~
    """

    #: These are the names (split by ::) for this namespace declaration,
    #: but does not include any parent namespace names
    #:
    #: An anonymous namespace is an empty list
    names: typing.List[str]
    inline: bool = False

    #: Documentation if present
    doxygen: typing.Optional[str] = None


@dataclass
class DecltypeSpecifier:
    """
    Contents of a decltype (inside the parentheses)

    .. code-block:: c++

        decltype(Foo::Bar)
                 ~~~~~~~~
    """

    #: Unparsed tokens within the decltype
    tokens: typing.List[Token]

    def format(self) -> str:
        return f"decltype({tokfmt(self.tokens)})"


@dataclass
class FundamentalSpecifier:
    """
    A specifier that only contains fundamental types.

    Fundamental types include various combinations of the following: unsigned,
    signed, short, int, long, float, double, char, bool, char16_t, char32_t,
    nullptr_t, wchar_t, void
    """

    name: str

    def format(self) -> str:
        return self.name


@dataclass
class NameSpecifier:
    """
    An individual segment of a type name

    .. code-block:: c++

        Foo::Bar
        ~~~

    """

    name: str

    specialization: typing.Optional["TemplateSpecialization"] = None

    def format(self) -> str:
        if self.specialization:
            return f"{self.name}{self.specialization.format()}"
        else:
            return self.name


@dataclass
class AutoSpecifier:
    """
    Used for an auto return type
    """

    name: str = "auto"

    def format(self) -> str:
        return self.name


@dataclass
class AnonymousName:
    """
    A name for an anonymous class, such as in a typedef. There is no string
    associated with this name, only an integer id. Things that share the same
    anonymous name have anonymous name instances with the same id
    """

    #: Unique id associated with this name (only unique per parser instance!)
    id: int

    def format(self) -> str:
        # TODO: not sure what makes sense here, subject to change
        return f"<<id={self.id}>>"


PQNameSegment = typing.Union[
    AnonymousName, FundamentalSpecifier, NameSpecifier, DecltypeSpecifier, AutoSpecifier
]


@dataclass
class PQName:
    """
    Possibly qualified name of a C++ type.
    """

    #: All of the segments of the name. This is always guaranteed to have at
    #: least one element in it. Name is segmented by '::'
    #:
    #: If a name refers to the global namespace, the first segment will be an
    #: empty NameSpecifier
    segments: typing.List[PQNameSegment]

    #: Set if the name starts with class/enum/struct
    classkey: typing.Optional[str] = None

    #: Set to true if the type was preceded with 'typename'
    has_typename: bool = False

    def format(self) -> str:
        tn = "typename " if self.has_typename else ""
        if self.classkey:
            return f"{tn}{self.classkey} {'::'.join(seg.format() for seg in self.segments)}"
        else:
            return tn + "::".join(seg.format() for seg in self.segments)


@dataclass
class TemplateArgument:
    """
    A single argument for a template specialization

    .. code-block:: c++

        Foo<int, Bar...>
            ~~~

    """

    #: If this argument is a type, it is stored here as a DecoratedType,
    #: otherwise it's stored as an unparsed set of values
    arg: typing.Union["DecoratedType", "FunctionType", Value]

    param_pack: bool = False

    def format(self) -> str:
        if self.param_pack:
            return f"{self.arg.format()}..."
        else:
            return self.arg.format()


@dataclass
class TemplateSpecialization:
    """
    Contains the arguments of a template specialization

    .. code-block:: c++

        Foo<int, Bar...>
            ~~~~~~~~~~~

    """

    args: typing.List[TemplateArgument]

    def format(self) -> str:
        return f"<{', '.join(arg.format() for arg in self.args)}>"


@dataclass
class FunctionType:
    """
    A function type, currently only used in a function pointer

    .. note:: There can only be one of FunctionType or Type in a DecoratedType
              chain
    """

    return_type: "DecoratedType"
    parameters: typing.List["Parameter"]

    #: If a member function pointer
    # TODO classname: typing.Optional[PQName]

    #: Set to True if ends with ``...``
    vararg: bool = False

    #: True if function has a trailing return type (``auto foo() -> int``).
    #: In this case, the 'auto' return type is removed and replaced with
    #: whatever the trailing return type was
    has_trailing_return: bool = False

    noexcept: typing.Optional[Value] = None

    #: Only set if an MSVC calling convention (__stdcall, etc) is explictly
    #: specified.
    #:
    #: .. note::  If your code contains things like WINAPI, you will need to
    #:            use a preprocessor to transform it to the appropriate
    #:            calling convention
    msvc_convention: typing.Optional[str] = None

    def format(self) -> str:
        vararg = "..." if self.vararg else ""
        params = ", ".join(p.format() for p in self.parameters)
        if self.has_trailing_return:
            return f"auto ({params}{vararg}) -> {self.return_type.format()}"
        else:
            return f"{self.return_type.format()} ({params}{vararg})"

    def format_decl(self, name: str) -> str:
        """Format as a named declaration"""
        vararg = "..." if self.vararg else ""
        params = ", ".join(p.format() for p in self.parameters)
        if self.has_trailing_return:
            return f"auto {name}({params}{vararg}) -> {self.return_type.format()}"
        else:
            return f"{self.return_type.format()} {name}({params}{vararg})"


@dataclass
class Type:
    """
    A type with a name associated with it
    """

    typename: PQName

    const: bool = False
    volatile: bool = False

    def format(self) -> str:
        c = "const " if self.const else ""
        v = "volatile " if self.volatile else ""
        return f"{c}{v}{self.typename.format()}"

    def format_decl(self, name: str):
        """Format as a named declaration"""
        c = "const " if self.const else ""
        v = "volatile " if self.volatile else ""
        return f"{c}{v}{self.typename.format()} {name}"


@dataclass
class Array:
    """
    Information about an array. Multidimensional arrays are represented as
    an array of array.
    """

    #: The type that this is an array of
    array_of: typing.Union["Array", "Pointer", Type]

    #: Size of the array
    #:
    #: .. code-block:: c++
    #:
    #:    int x[10];
    #:          ~~
    size: typing.Optional[Value]

    def format(self) -> str:
        s = self.size.format() if self.size else ""
        return f"{self.array_of.format()}[{s}]"

    def format_decl(self, name: str) -> str:
        s = self.size.format() if self.size else ""
        return f"{self.array_of.format()} {name}[{s}]"


@dataclass
class Pointer:
    """
    A pointer
    """

    #: Thing that this points to
    ptr_to: typing.Union[Array, FunctionType, "Pointer", Type]

    const: bool = False
    volatile: bool = False
    restrict: bool = False

    def format(self) -> str:
        c = " const" if self.const else ""
        v = " volatile" if self.volatile else ""
        r = " __restrict__" if self.restrict else ""
        ptr_to = self.ptr_to
        if isinstance(ptr_to, (Array, FunctionType)):
            return ptr_to.format_decl(f"(*{r}{c}{v})")
        else:
            return f"{ptr_to.format()}*{r}{c}{v}"

    def format_decl(self, name: str):
        """Format as a named declaration"""
        c = " const" if self.const else ""
        v = " volatile" if self.volatile else ""
        r = " __restrict__" if self.restrict else ""
        ptr_to = self.ptr_to
        if isinstance(ptr_to, (Array, FunctionType)):
            return ptr_to.format_decl(f"(*{r}{c}{v} {name})")
        else:
            return f"{ptr_to.format()}*{r}{c}{v} {name}"


@dataclass
class Reference:
    """
    A lvalue (``&``) reference
    """

    ref_to: typing.Union[Array, FunctionType, Pointer, Type]
    restrict: bool = False

    def format(self) -> str:
        ref_to = self.ref_to

        if isinstance(ref_to, Array):
            return ref_to.format_decl("(&)")
        else:
            r = " __restrict__" if self.restrict else ""
            return f"{ref_to.format()}&{r}"

    def format_decl(self, name: str):
        """Format as a named declaration"""
        ref_to = self.ref_to

        if isinstance(ref_to, Array):
            return ref_to.format_decl(f"(& {name})")
        else:
            r = " __restrict__" if self.restrict else ""
            return f"{ref_to.format()}&{r} {name}"


@dataclass
class MoveReference:
    """
    An rvalue (``&&``) reference
    """

    moveref_to: typing.Union[Array, FunctionType, Pointer, Type]

    def format(self) -> str:
        return f"{self.moveref_to.format()}&&"

    def format_decl(self, name: str):
        """Format as a named declaration"""
        return f"{self.moveref_to.format()}&& {name}"


#: A type or function type that is decorated with various things
#:
#: .. note:: There can only be one of FunctionType or Type in a DecoratedType
#:           chain
DecoratedType = typing.Union[Array, Pointer, MoveReference, Reference, Type]


@dataclass
class Enumerator:
    """
    An individual value of an enumeration
    """

    #: The enumerator key name
    name: str

    #: None if not explicitly specified
    value: typing.Optional[Value] = None

    #: Documentation if present
    doxygen: typing.Optional[str] = None

    #: Any attributes attached to this enumerator
    attributes: typing.List[Attribute] = field(default_factory=list)


@dataclass
class EnumDecl:
    """
    An enumeration type
    """

    typename: PQName

    values: typing.List[Enumerator]

    base: typing.Optional[PQName] = None

    #: Documentation if present
    doxygen: typing.Optional[str] = None
    #: Any attributes attached to this enum declaration
    attributes: typing.List[Attribute] = field(default_factory=list)

    #: If within a class, the access level for this decl
    access: typing.Optional[str] = None


@dataclass
class TemplateNonTypeParam:
    """

    .. code-block:: c++

       template <int T>
                 ~~~~~

       template <class T, typename T::type* U>
                          ~~~~~~~~~~~~~~~~~~~

       template <auto T>
                 ~~~~~~

       // abbreviated template parameters are converted to this and param_idx is set
       void fn(C auto p)
               ~~~~~~

       // abbreviated template parameters that are return types have param_idx = -1
       C auto fn()
       ~~~~~~
    """

    type: DecoratedType
    name: typing.Optional[str] = None
    default: typing.Optional[Value] = None

    #: If this was promoted, the parameter index that this corresponds with. Return
    #: types are set to -1
    param_idx: typing.Optional[int] = None

    #: Contains a ``...``
    param_pack: bool = False


@dataclass
class TemplateTypeParam:
    """

    .. code-block:: c++

       template <typename T>
                 ~~~~~~~~~~
    """

    #: 'typename' or 'class'
    typekey: str

    name: typing.Optional[str] = None

    param_pack: bool = False

    default: typing.Optional[Value] = None

    #: A template-template param
    template: typing.Optional["TemplateDecl"] = None


#: A parameter for a template declaration
#:
#: .. code-block:: c++
#:
#:    template <typename T>
#:              ~~~~~~~~~~
TemplateParam = typing.Union[TemplateNonTypeParam, TemplateTypeParam]


@dataclass
class TemplateDecl:
    """
    Template declaration for a function or class

    .. code-block:: c++

        template <typename T>
        class Foo {};

        template <typename T>
        T fn();

    """

    params: typing.List[TemplateParam] = field(default_factory=list)

    # Currently don't interpret requires, if that changes in the future
    # then this API will change.

    #: template <typename T> requires ...
    raw_requires_pre: typing.Optional[Value] = None


#: If no template, this is None. This is a TemplateDecl if this there is a single
#: declaration:
#:
#: .. code-block:: c++
#:
#:    template <typename T>
#:    struct C {};
#:
#: If there are multiple template declarations, then this is a list of
#: declarations in the order that they're encountered:
#:
#: .. code-block:: c++
#:
#:    template<>
#:    template<class U>
#:    struct A<char>::C {};
#:
TemplateDeclTypeVar = typing.Union[None, TemplateDecl, typing.List[TemplateDecl]]


@dataclass
class TemplateInst:
    """
    Explicit template instantiation

    .. code-block:: c++

        template class MyClass<1,2>;

        extern template class MyClass<2,3>;
    """

    typename: PQName
    extern: bool
    doxygen: typing.Optional[str] = None


@dataclass
class Concept:
    """
    Preliminary support for consuming headers that contain concepts, but
    not trying to actually make sense of them at this time. If this is
    something you care about, pull requests are welcomed!

    .. code-block:: c++

        template <class T>
        concept Meowable = is_meowable<T>;

        template<typename T>
        concept Addable = requires (T x) { x + x; };
    """

    template: TemplateDecl
    name: str

    #: In the future this will be removed if we fully parse the expression
    raw_constraint: Value

    doxygen: typing.Optional[str] = None


@dataclass
class ForwardDecl:
    """
    Represents a forward declaration of a user defined type
    """

    typename: PQName
    template: TemplateDeclTypeVar = None
    doxygen: typing.Optional[str] = None
    #: Any attributes attached to this forward declaration
    attributes: typing.List[Attribute] = field(default_factory=list)

    #: Set if this is a forward declaration of an enum and it has a base
    enum_base: typing.Optional[PQName] = None

    #: If within a class, the access level for this decl
    access: typing.Optional[str] = None


@dataclass
class BaseClass:
    """
    Base class declarations for a class
    """

    #: access specifier for this base
    access: str

    #: possibly qualified type name for the base
    typename: PQName

    #: Virtual inheritance
    virtual: bool = False

    #: Contains a ``...``
    param_pack: bool = False
    #: Any attributes attached to this base specifier
    attributes: typing.List[Attribute] = field(default_factory=list)


@dataclass
class ClassDecl:
    """
    A class is a user defined type (class/struct/union)
    """

    typename: PQName

    bases: typing.List[BaseClass] = field(default_factory=list)
    template: TemplateDeclTypeVar = None

    explicit: bool = False
    final: bool = False

    doxygen: typing.Optional[str] = None
    #: Any attributes attached to this class declaration
    attributes: typing.List[Attribute] = field(default_factory=list)

    #: If within a class, the access level for this decl
    access: typing.Optional[str] = None

    @property
    def classkey(self) -> typing.Optional[str]:
        return self.typename.classkey


@dataclass
class Parameter:
    """
    A parameter of a function/method
    """

    type: DecoratedType
    name: typing.Optional[str] = None
    default: typing.Optional[Value] = None
    param_pack: bool = False

    #: Indicates that this parameter is the parameter used for cxx23 deducing-this
    deduces_this: bool = False

    def format(self) -> str:
        default = f" = {self.default.format()}" if self.default else ""
        pp = "... " if self.param_pack else ""
        name = self.name
        if name:
            return f"{self.type.format_decl(f'{pp}{name}')}{default}"
        else:
            return f"{self.type.format()}{pp}{default}"


@dataclass
class Function:
    """
    A function declaration, potentially with the function body
    """

    #: Only constructors and destructors don't have a return type
    return_type: typing.Optional[DecoratedType]

    name: PQName
    parameters: typing.List[Parameter]

    #: Set to True if ends with ``...``
    vararg: bool = False

    doxygen: typing.Optional[str] = None
    #: Any attributes attached to this function declaration
    attributes: typing.List[Attribute] = field(default_factory=list)

    constexpr: bool = False
    extern: typing.Union[bool, str] = False
    static: bool = False
    inline: bool = False
    deleted: bool = False

    #: If true, the body of the function is present
    has_body: bool = False

    #: True if function has a trailing return type (``auto foo() -> int``).
    #: In this case, the 'auto' return type is removed and replaced with
    #: whatever the trailing return type was
    has_trailing_return: bool = False

    template: TemplateDeclTypeVar = None

    #: Value of any throw specification for this function. The value omits the
    #: outer parentheses.
    throw: typing.Optional[Value] = None

    #: Value of any noexcept specification for this function. The value omits
    #: the outer parentheses.
    noexcept: typing.Optional[Value] = None

    #: Only set if an MSVC calling convention (__stdcall, etc) is explictly
    #: specified.
    #:
    #: .. note::  If your code contains things like WINAPI, you will need to
    #:            use a preprocessor to transform it to the appropriate
    #:            calling convention
    msvc_convention: typing.Optional[str] = None

    #: The operator type (+, +=, etc).
    #:
    #: If this object is a Function, then this is a free operator function. If
    #: this object is a Method, then it is an operator method.
    #:
    #: In the case of a conversion operator (such as 'operator bool'), this
    #: is the string "conversion" and the full Type is found in return_type
    operator: typing.Optional[str] = None

    #: A requires constraint following the function declaration. If you need the
    #: prior, look at TemplateDecl.raw_requires_pre. At the moment this is just
    #: a raw value, if we interpret it in the future this will change.
    #:
    #: template <typename T> int main() requires ...
    raw_requires: typing.Optional[Value] = None


@dataclass
class Method(Function):
    """
    A method declaration, potentially with the method body
    """

    #: If parsed within a class, the access level for this method
    access: typing.Optional[str] = None

    const: bool = False
    volatile: bool = False

    #: ref-qualifier for this method, either lvalue (&) or rvalue (&&)
    #:
    #: .. code-block:: c++
    #:
    #:   void foo() &&;
    #:              ~~
    #:
    ref_qualifier: typing.Optional[str] = None

    constructor: bool = False

    #: True if the method was declared ``explicit``. For C++20
    #: ``explicit(<expr>)``, this holds the expression as a Value (omitting
    #: the outer parentheses).
    explicit: typing.Union[bool, Value] = False
    default: bool = False

    destructor: bool = False

    pure_virtual: bool = False
    virtual: bool = False
    final: bool = False
    override: bool = False


@dataclass
class FriendDecl:
    """
    Represents a friend declaration -- friends can only be classes or functions
    """

    cls: typing.Optional[ForwardDecl] = None

    fn: typing.Optional[Function] = None


@dataclass
class Typedef:
    """
    A typedef specifier. A unique typedef specifier is created for each alias
    created by the typedef.

    .. code-block:: c++

        typedef type name, *pname;

    """

    #: The aliased type or function type
    #:
    #: .. code-block:: c++
    #:
    #:    typedef type *pname;
    #:            ~~~~~~
    type: typing.Union[DecoratedType, FunctionType]

    #: The alias introduced for the specified type
    #:
    #: .. code-block:: c++
    #:
    #:    typedef type *pname;
    #:                  ~~~~~
    name: str

    #: If within a class, the access level for this decl
    access: typing.Optional[str] = None
    #: Any attributes attached to this typedef
    attributes: typing.List[Attribute] = field(default_factory=list)


@dataclass
class Variable:
    """
    A variable declaration
    """

    name: PQName
    type: DecoratedType

    value: typing.Optional[Value] = None

    constexpr: bool = False
    extern: typing.Union[bool, str] = False
    static: bool = False
    inline: bool = False

    #: Can occur for a static variable for a templated class
    template: typing.Optional[TemplateDecl] = None

    doxygen: typing.Optional[str] = None
    #: Any attributes attached to this variable
    attributes: typing.List[Attribute] = field(default_factory=list)


@dataclass
class Field:
    """
    A field of a class
    """

    #: public/private/protected
    access: str

    type: DecoratedType
    name: typing.Optional[str] = None

    value: typing.Optional[Value] = None
    bits: typing.Union[None, int, Value] = None

    constexpr: bool = False
    mutable: bool = False
    static: bool = False
    inline: bool = False

    doxygen: typing.Optional[str] = None
    #: Any attributes attached to this field
    attributes: typing.List[Attribute] = field(default_factory=list)


@dataclass
class UsingDecl:
    """
    .. code-block:: c++

        using NS::ClassName;
    """

    typename: PQName

    #: If within a class, the access level for this decl
    access: typing.Optional[str] = None

    #: Documentation if present
    doxygen: typing.Optional[str] = None


@dataclass
class UsingAlias:
    """
    .. code-block:: c++

        using foo = int;

        template <typename T>
        using VectorT = std::vector<T>;

    """

    alias: str
    type: DecoratedType

    template: typing.Optional[TemplateDecl] = None

    #: If within a class, the access level for this decl
    access: typing.Optional[str] = None

    #: Documentation if present
    doxygen: typing.Optional[str] = None


@dataclass
class DeductionGuide:
    """
    .. code-block:: c++

    template <class T>
    MyClass(T) -> MyClass(int);
    """

    #: Only constructors and destructors don't have a return type
    result_type: typing.Optional[DecoratedType]

    name: PQName
    parameters: typing.List[Parameter]

    doxygen: typing.Optional[str] = None



import typing

if typing.TYPE_CHECKING:
    from .lexer import LexToken


class CxxParseError(Exception):
    """
    Exception raised when a parsing error occurs
    """

    def __init__(self, msg: str, tok: typing.Optional["LexToken"] = None) -> None:
        Exception.__init__(self, msg)
        self.tok = tok



import argparse
import dataclasses
import json
import pathlib
import pprint
import subprocess
import sys

from .options import ParserOptions
from .simple import parse_file


def dumpmain() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("header")
    parser.add_argument(
        "-w",
        "--width",
        default=80,
        type=int,
        help="Width of output when in pprint mode",
    )
    parser.add_argument("-v", "--verbose", default=False, action="store_true")
    parser.add_argument(
        "--mode",
        choices=["json", "pprint", "repr", "brepr", "pponly"],
        default="pprint",
    )
    parser.add_argument(
        "--pcpp", default=False, action="store_true", help="Use pcpp preprocessor"
    )
    parser.add_argument(
        "--gcc", default=False, action="store_true", help="Use GCC as preprocessor"
    )
    parser.add_argument(
        "--depfile",
        default=None,
        type=pathlib.Path,
        help="Generate a depfile (requires preprocessor)",
    )
    parser.add_argument(
        "--deptarget", default=[], action="append", help="depfile target"
    )
    parser.add_argument(
        "--encoding", default=None, help="Use this encoding to open the file"
    )

    args = parser.parse_args()

    pp_kwargs = dict(encoding=args.encoding)

    if args.depfile:
        if not (args.pcpp or args.gcc):
            parser.error("--depfile requires either --pcpp or --gcc")

        pp_kwargs["depfile"] = args.depfile
        pp_kwargs["deptarget"] = args.deptarget

    preprocessor = None
    if args.gcc:
        from .preprocessor import make_gcc_preprocessor

        preprocessor = make_gcc_preprocessor(**pp_kwargs)

    if args.pcpp or (args.mode == "pponly" and preprocessor is None):
        from .preprocessor import make_pcpp_preprocessor

        preprocessor = make_pcpp_preprocessor(**pp_kwargs)

    if args.mode == "pponly":
        assert preprocessor is not None
        with open(args.header, "r", encoding=args.encoding) as fp:
            pp_content = preprocessor(args.header, fp.read())
        sys.stdout.write(pp_content)
        sys.exit(0)

    options = ParserOptions(verbose=args.verbose, preprocessor=preprocessor)
    data = parse_file(args.header, encoding=args.encoding, options=options)

    if args.mode == "pprint":
        ddata = dataclasses.asdict(data)
        pprint.pprint(ddata, width=args.width, compact=True)

    elif args.mode == "json":
        ddata = dataclasses.asdict(data)
        json.dump(ddata, sys.stdout, indent=2)

    elif args.mode == "brepr":
        stmt = repr(data)
        stmt = subprocess.check_output(
            ["black", "-", "-q"], input=stmt.encode("utf-8")
        ).decode("utf-8")

        print(stmt)

    elif args.mode == "repr":
        print(data)

    else:
        parser.error("Invalid mode")



import sys
import typing

if sys.version_info >= (3, 8):
    from typing import Protocol
else:
    Protocol = object  # pragma: no cover


from .types import (
    Concept,
    DeductionGuide,
    EnumDecl,
    Field,
    ForwardDecl,
    FriendDecl,
    Function,
    Method,
    NamespaceAlias,
    TemplateInst,
    Typedef,
    UsingAlias,
    UsingDecl,
    Variable,
    Value,
)

from .parserstate import (
    State,
    ClassBlockState,
    ExternBlockState,
    NamespaceBlockState,
    NonClassBlockState,
)


class CxxVisitor(Protocol):
    """
    Defines the interface used by the parser to emit events
    """

    def on_parse_start(self, state: NamespaceBlockState) -> None:
        """
        Called when parsing begins
        """

    def on_pragma(self, state: State, content: Value) -> None:
        """
        Called once for each ``#pragma`` directive encountered
        """

    def on_include(self, state: State, filename: str) -> None:
        """
        Called once for each ``#include`` directive encountered
        """

    def on_extern_block_start(self, state: ExternBlockState) -> typing.Optional[bool]:
        """
        .. code-block:: c++

            extern "C" {

            }

        If this function returns False, the visitor will not be called for any
        items inside this block (including on_extern_block_end)
        """

    def on_extern_block_end(self, state: ExternBlockState) -> None:
        """
        Called when an extern block ends
        """

    def on_namespace_start(self, state: NamespaceBlockState) -> typing.Optional[bool]:
        """
        Called when a ``namespace`` directive is encountered

        If this function returns False, the visitor will not be called for any
        items inside this namespace (including on_namespace_end)
        """

    def on_namespace_end(self, state: NamespaceBlockState) -> None:
        """
        Called at the end of a ``namespace`` block
        """

    def on_namespace_alias(
        self, state: NonClassBlockState, alias: NamespaceAlias
    ) -> None:
        """
        Called when a ``namespace`` alias is encountered
        """

    def on_concept(self, state: NonClassBlockState, concept: Concept) -> None:
        """
        .. code-block:: c++

            template <class T>
            concept Meowable = is_meowable<T>;
        """

    def on_forward_decl(self, state: State, fdecl: ForwardDecl) -> None:
        """
        Called when a forward declaration is encountered
        """

    def on_template_inst(self, state: State, inst: TemplateInst) -> None:
        """
        Called when an explicit template instantiation is encountered
        """

    def on_variable(self, state: State, v: Variable) -> None:
        """
        Called when a global variable is encountered
        """

    def on_function(self, state: NonClassBlockState, fn: Function) -> None:
        """
        Called when a function is encountered that isn't part of a class
        """

    def on_method_impl(self, state: NonClassBlockState, method: Method) -> None:
        """
        Called when a method implementation is encountered outside of a class
        declaration. For example:

        .. code-block:: c++

            void MyClass::fn() {
                // does something
            }

        .. note:: The above implementation is ambiguous, as it technically could
                  be a function in a namespace. We emit this instead as it's
                  more likely to be the case in common code.
        """

    def on_typedef(self, state: State, typedef: Typedef) -> None:
        """
        Called for each typedef instance encountered. For example:

        .. code-block:: c++

            typedef int T, *PT;

        Will result in ``on_typedef`` being called twice, once for ``T`` and
        once for ``*PT``
        """

    def on_using_namespace(
        self, state: NonClassBlockState, namespace: typing.List[str]
    ) -> None:
        """
        .. code-block:: c++

            using namespace std;
        """

    def on_using_alias(self, state: State, using: UsingAlias) -> None:
        """
        .. code-block:: c++

            using foo = int;

            template <typename T>
            using VectorT = std::vector<T>;

        """

    def on_using_declaration(self, state: State, using: UsingDecl) -> None:
        """
        .. code-block:: c++

            using NS::ClassName;

        """

    #
    # Enums
    #

    def on_enum(self, state: State, enum: EnumDecl) -> None:
        """
        Called after an enum is encountered
        """

    #
    # Class/union/struct
    #

    def on_class_start(self, state: ClassBlockState) -> typing.Optional[bool]:
        """
        Called when a class/struct/union is encountered

        When part of a typedef:

        .. code-block:: c++

            typedef struct { } X;

        This is called first, followed by on_typedef for each typedef instance
        encountered. The compound type object is passed as the type to the
        typedef.

        If this function returns False, the visitor will not be called for any
        items inside this class (including on_class_end)
        """

    def on_class_field(self, state: ClassBlockState, f: Field) -> None:
        """
        Called when a field of a class is encountered
        """

    def on_class_friend(self, state: ClassBlockState, friend: FriendDecl) -> None:
        """
        Called when a friend declaration is encountered
        """

    def on_class_method(self, state: ClassBlockState, method: Method) -> None:
        """
        Called when a method of a class is encountered inside of a class
        """

    def on_class_end(self, state: ClassBlockState) -> None:
        """
        Called when the end of a class/struct/union is encountered.

        When a variable like this is declared:

        .. code-block:: c++

            struct X {

            } x;

        Then ``on_class_start``, .. ``on_class_end`` are emitted, along with
        ``on_variable`` for each instance declared.
        """

    def on_deduction_guide(
        self, state: NonClassBlockState, guide: DeductionGuide
    ) -> None:
        """
        Called when a deduction guide is encountered
        """


class NullVisitor:
    """
    This visitor does nothing
    """

    def on_parse_start(self, state: NamespaceBlockState) -> None:
        return None

    def on_pragma(self, state: State, content: Value) -> None:
        return None

    def on_include(self, state: State, filename: str) -> None:
        return None

    def on_extern_block_start(self, state: ExternBlockState) -> typing.Optional[bool]:
        return None

    def on_extern_block_end(self, state: ExternBlockState) -> None:
        return None

    def on_namespace_start(self, state: NamespaceBlockState) -> typing.Optional[bool]:
        return None

    def on_namespace_end(self, state: NamespaceBlockState) -> None:
        return None

    def on_concept(self, state: NonClassBlockState, concept: Concept) -> None:
        return None

    def on_namespace_alias(
        self, state: NonClassBlockState, alias: NamespaceAlias
    ) -> None:
        return None

    def on_forward_decl(self, state: State, fdecl: ForwardDecl) -> None:
        return None

    def on_template_inst(self, state: State, inst: TemplateInst) -> None:
        return None

    def on_variable(self, state: State, v: Variable) -> None:
        return None

    def on_function(self, state: NonClassBlockState, fn: Function) -> None:
        return None

    def on_method_impl(self, state: NonClassBlockState, method: Method) -> None:
        return None

    def on_typedef(self, state: State, typedef: Typedef) -> None:
        return None

    def on_using_namespace(
        self, state: NonClassBlockState, namespace: typing.List[str]
    ) -> None:
        return None

    def on_using_alias(self, state: State, using: UsingAlias) -> None:
        return None

    def on_using_declaration(self, state: State, using: UsingDecl) -> None:
        return None

    def on_enum(self, state: State, enum: EnumDecl) -> None:
        return None

    def on_class_start(self, state: ClassBlockState) -> typing.Optional[bool]:
        return None

    def on_class_field(self, state: ClassBlockState, f: Field) -> None:
        return None

    def on_class_friend(self, state: ClassBlockState, friend: FriendDecl) -> None:
        return None

    def on_class_method(self, state: ClassBlockState, method: Method) -> None:
        return None

    def on_class_end(self, state: ClassBlockState) -> None:
        return None

    def on_deduction_guide(
        self, state: NonClassBlockState, guide: DeductionGuide
    ) -> None:
        return None


null_visitor = NullVisitor()



import argparse
import dataclasses
import enum
import inspect
import subprocess
import textwrap
import typing

from .errors import CxxParseError
from .preprocessor import make_pcpp_preprocessor
from .options import ParserOptions
from .simple import parse_string, parse_typename, ParsedData


def nondefault_repr(data: typing.Any) -> str:
    """
    Similar to the default dataclass repr, but exclude any
    default parameters or parameters with compare=False
    """

    is_dataclass = dataclasses.is_dataclass
    get_fields = dataclasses.fields
    MISSING = dataclasses.MISSING

    def _inner_repr(o: typing.Any) -> str:
        if is_dataclass(o):
            vals = []
            for f in get_fields(o):
                if f.repr and f.compare:
                    v = getattr(o, f.name)
                    if f.default_factory is not MISSING:
                        default = f.default_factory()
                    else:
                        default = f.default

                    if v != default:
                        vals.append(f"{f.name}={_inner_repr(v)}")

            return f"{o.__class__.__qualname__ }({', '.join(vals)})"

        elif isinstance(o, list):
            return f"[{','.join(_inner_repr(l) for l in o)}]"
        elif isinstance(o, dict):
            vals = []
            for k, v in o.items():
                vals.append(f'"{k}": {_inner_repr(v)}')
            return "{" + ",".join(vals) + "}"
        elif isinstance(o, enum.Enum):
            return f"{o.__class__.__qualname__}.{o.name}"
        else:
            return repr(o)

    return _inner_repr(data)


def gentest(
    infile: str,
    name: str,
    outfile: str,
    verbose: bool,
    fail: bool,
    pcpp: bool,
    typename_mode: bool,
) -> None:
    # Goal is to allow making a unit test as easy as running this dumper
    # on a file and copy/pasting this into a test

    with open(infile, "r") as fp:
        content = fp.read()

    maybe_options = ""
    popt = ""

    options = ParserOptions(verbose=verbose)
    if pcpp:
        options.preprocessor = make_pcpp_preprocessor()
        maybe_options = "options = ParserOptions(preprocessor=make_pcpp_preprocessor())"
        popt = ", options=options"

    if typename_mode:
        try:
            dtype = parse_typename(content.strip(), options=options)
            if fail:
                raise ValueError("did not fail")
        except CxxParseError:
            if not fail:
                raise
            try:
                parse_typename(content.strip(), options=options)
            except CxxParseError as e2:
                err = str(e2)

        if not fail:
            stmt = nondefault_repr(dtype)
            stmt = f"""
                {maybe_options}
                dtype = parse_typename(content.strip(){popt})

                assert dtype == {stmt}
            """
        else:
            stmt = f"""
                {maybe_options}
                err = {repr(err)}
                with pytest.raises(CxxParseError, match=re.escape(err)):
                    parse_typename(content.strip(){popt})
            """
    else:
        try:
            data = parse_string(content, options=options)
            if fail:
                raise ValueError("did not fail")
        except CxxParseError:
            if not fail:
                raise
            # do it again, but strip the content so the error message matches
            try:
                parse_string(content.strip(), options=options)
            except CxxParseError as e2:
                err = str(e2)

        if not fail:
            stmt = nondefault_repr(data)
            stmt = f"""
                {maybe_options}
                data = parse_string(content, cleandoc=True{popt})

                assert data == {stmt}
            """
        else:
            stmt = f"""
                {maybe_options}
                err = {repr(err)}
                with pytest.raises(CxxParseError, match=re.escape(err)):
                    parse_string(content, cleandoc=True{popt})
            """

    stmt = textwrap.dedent(stmt).strip()
    stmt = textwrap.indent(stmt, " " * 4)
    content = inspect.cleandoc(content)
    content = textwrap.indent("\n" + content, " " * 8)
    content = "\n".join(l.rstrip() for l in content.splitlines())

    stmt = f"""def test_{name}() -> None:
    content = \"\"\"{content}
    \"\"\"

{stmt}
"""

    # format it with black
    stmt = subprocess.check_output(
        ["black", "-", "-q"], input=stmt.encode("utf-8")
    ).decode("utf-8")

    if outfile == "-":
        print(stmt)
    else:
        with open(outfile, "w") as fp:
            fp.write(stmt)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("header")
    parser.add_argument("name", nargs="?", default="TODO")
    parser.add_argument("--pcpp", default=False, action="store_true")
    parser.add_argument(
        "--typename",
        dest="typename_mode",
        default=False,
        action="store_true",
        help="Generate tests for parse_typename",
    )
    parser.add_argument("-v", "--verbose", default=False, action="store_true")
    parser.add_argument("-o", "--output", default="-")
    parser.add_argument(
        "-x", "--fail", default=False, action="store_true", help="Expect failure"
    )
    args = parser.parse_args()

    gentest(
        args.header,
        args.name,
        args.output,
        args.verbose,
        args.fail,
        args.pcpp,
        args.typename_mode,
    )



from dataclasses import dataclass
from typing import Callable, Optional

#: arguments are (filename, content)
PreprocessorFunction = Callable[[str, Optional[str]], str]


@dataclass
class ParserOptions:
    """
    Options that control parsing behaviors
    """

    #: If true, prints out
    verbose: bool = False

    #: If true, converts a single void parameter to zero parameters
    convert_void_to_zero_params: bool = True

    #: A function that will preprocess the header before parsing. See
    #: :py:mod:`cxxheaderparser.preprocessor` for available preprocessors
    preprocessor: Optional[PreprocessorFunction] = None
